@@ -16,17 +16,18 @@ from socket import (
     getaddrinfo,
     socket,
 )
-from threading import Thread, Timer
+from threading import Condition, Thread, Timer
 from time import monotonic, sleep
 from typing import Dict, Optional, Set
 from uuid import uuid4
 from wsgiref import simple_server
 
-from werkzeug.serving import WSGIRequestHandler
+from werkzeug.serving import WSGIRequestHandler, BaseWSGIServer, make_server
 
 from .gating import wait_for_idle_cpus
 from .oob_response import OutOfBandResponder
 from .resource_balancer import BoundedResourceBalancer, ResourceLoan
+from .subprocess_middleware import SubprocessMiddleware
 from .util import ObjectProxy, ThreadsTracker
 
 LOGGER = getLogger("pothead.worker")
@@ -50,6 +51,24 @@ class DeferrableCleanup:
             self.f()
 
 
+class Flag:
+    def __init__(self):
+        self._set = False
+        self._cond = Condition()
+
+    def __bool__(self):
+        return self._set
+
+    def set(self):
+        with self._cond:
+            self._set = True
+            self._cond.notify_all()
+
+    def wait(self, timeout: float):
+        with self._cond:
+            self._cond.wait_for(lambda: self._set, timeout)
+
+
 class Handler(WSGIRequestHandler):
     protocol_version = "HTTP/1.1"
 
@@ -92,7 +111,7 @@ class Handler(WSGIRequestHandler):
         success = super().parse_request()
         if success:
             self.request_received_at = monotonic()
-            self.request_id = self.headers.get('X-Request-Id', None) or str(uuid4())
+            self.request_id = self.headers.get("X-Request-Id", None) or str(uuid4())
 
         # This socket shall no longer be fast-closed on SIGTERM
         socket = self.request
@@ -115,7 +134,9 @@ class Handler(WSGIRequestHandler):
 
     def log_elapsed(self):
         LOGGER.info(
-            "[%s] Elapsed: %.3f", self.request_id, monotonic() - self.request_received_at
+            "[%s] Elapsed: %.3f",
+            self.request_id,
+            monotonic() - self.request_received_at,
         )
 
 
@@ -146,12 +167,14 @@ class LoadBalancer:
 
 
 class Server:
+    shutdown_signal = (
+        False  # Must exist for werkzeug 1.x legacy. Can be dropped with werkzeug 2.x
+    )
     ssl_context = None
     multithread = False
     multiprocess = False
     passthrough_errors = False
-    shutdown_signal = False
-    running = True
+    inspection_server: Optional[BaseWSGIServer] = None
 
     def __init__(
         self,
@@ -160,6 +183,8 @@ class Server:
         *,
         load_balancer=LoadBalancer,
         redirect_port: Optional[int] = None,
+        inspection_port: Optional[int] = None,
+        use_request_subprocess: bool = False,
         oob_cleanup_interval_seconds: Optional[float] = None,
         oob_item_expiry_seconds: Optional[float] = None,
     ):
@@ -176,21 +201,53 @@ class Server:
 
         self.broker_balancer = load_balancer(host, port)
         self.sockets_waiting_for_request: Set[socket] = set()
-        self.real_app = app
         self.worker_tracker = ThreadsTracker()
+
+        self.stopping = Flag()
+        self.stopped = Flag()
+
+        if inspection_port is not None:
+            self.inspection_server = make_server(
+                "", inspection_port, self.inspect, threaded=True
+            )
+            Thread(target=self.inspection_server.serve_forever, daemon=True).start()
+
+        self.wait_for_slot = app.wait_for_slot
+        del app.wait_for_slot
+        self.app = app
+        self.app = (
+            SubprocessMiddleware(self.app) if use_request_subprocess else self.app
+        )
         self.app = (
             OutOfBandResponder(
-                app,
+                self.app,
                 redirect_port,
                 oob_cleanup_interval_seconds,
                 oob_item_expiry_seconds,
             )
             if redirect_port is not None
-            else app
+            else self.app
         )
 
+    def inspect(self, environ, start_response):
+        if environ.get("REQUEST_METHOD", None) != "GET":
+            start_response("405 Only GET supported", ())
+            return
+        path = environ.get("PATH_INFO", None)
+        if path == "/inspect":
+            start_response("200 OK", ())
+            yield f"{len(self.worker_tracker)} workers running\n".encode()
+            yield f"{len(self.sockets_waiting_for_request)} blocked waiting for request\n".encode()
+        elif path == "/apocalypse":
+            start_response("200 OK", ())
+            while not self.stopped.wait(5):
+                yield b"Not yet\n"
+            yield b"It's time\n"
+        else:
+            start_response("404 Not found", ())
+
     def worker(self):
-        while self.running:
+        while not self.stopping:
             with self.worker_tracker, self.broker_balancer.acquire_addr() as addr:
                 self._fetch_and_run_one_job(addr)
 
@@ -234,7 +291,7 @@ class Server:
 
     def shutdown(self):
         LOGGER.info("Shutting down workers...")
-        self.running = False
+        self.stopping.set()
         for s in list(self.sockets_waiting_for_request):
             try:
                 s.shutdown(SHUT_RDWR)
@@ -247,6 +304,10 @@ class Server:
         if isinstance(self.app, OutOfBandResponder):
             self.app.shutdown()
         LOGGER.info("Worker shutdown completed")
+        self.stopped.set()
+        if self.inspection_server is not None:
+            self.inspection_server.shutdown()
+        LOGGER.info("All services stopped")
 
     def poll_loop(self):
         """Uses app.wait_for_slot() spawn new workers dynamically. This is useful, for example to wait for
@@ -261,14 +322,14 @@ class Server:
 
         See pothead.gating for some ready-made `wait_for_slot` implementations"""
 
-        while self.running:
-            callbacks = self.real_app.wait_for_slot(lambda: not self.running)
-            if not self.running:
+        while not self.stopping:
+            callbacks = self.wait_for_slot(lambda: bool(self.stopping))
+            if self.stopping:
                 break
 
             # This will block as long as we are waiting for requests from all broker-instances
             broker_instance_lease = self.broker_balancer.acquire_addr()
-            if not self.running:
+            if self.stopping:
                 break
 
             Thread(
@@ -338,6 +399,21 @@ if __name__ == "__main__":
         default=None,
         help="When streaming successful responses, first redirect client to worker at [port]",
     )
+    parser.add_argument(
+        "--inspection-port",
+        default=None,
+        type=int,
+        help="If set, the server will serve a small inspection-API on [port]",
+    )
+    parser.add_argument(
+        "--use-request-subprocess",
+        action="store_true",
+        help=(
+            "Spawn a separate WSGI app subprocess for handling each request. "
+            + "Note: Only suitable for use cases where the request payload is small."
+        ),
+    )
+
     job_control_group = parser.add_mutually_exclusive_group()
     job_control_group.add_argument(
         "--workers",
@@ -363,7 +439,13 @@ if __name__ == "__main__":
 
     LOGGER.info("Initializing server for app %s", args.app)
 
-    server = Server(args.connect, args.app, redirect_port=args.redirect_response)
+    server = Server(
+        args.connect,
+        args.app,
+        redirect_port=args.redirect_response,
+        use_request_subprocess=args.use_request_subprocess,
+        inspection_port=args.inspection_port,
+    )
 
     install_term_handler(server.shutdown)
 
